----------------------------------------------------------------------------
-- The Standardisation Theorem for weak-head/internal reduction

{-# OPTIONS --without-K --safe #-}

module standardisation where

open import Data.Nat.Base using (suc)
open import Data.Fin.Base using (Fin)
open import Data.Vec.Base using ()
  renaming ([] to ε; lookup to _⟦_⟧)
open import Data.Vec.Relation.Binary.Pointwise.Inductive
  renaming ([] to ε∼ε; refl to [_]; lookup to _at_)
open import Data.Fin.Substitution.Example as 𝔼
  renaming (Tm to Lam; var to [_]ᵥ; _·_ to _•_;
            _/_ to _/ₛ_; _↑ to _↑ₛ; _↑⋆_ to _↑ₛ⋆_;
            wk to wkₛ; sub to subₛ; id to idₛ)
open import Data.Fin.Substitution as 𝕊
open import Level renaming (zero to ℓ₀) hiding (suc)
open import Relation.Binary using (Rel)
open import Relation.Binary.PropositionalEquality as ℙ
  hiding ([_])
open ℙ.≡-Reasoning


---------------------------------------------------------------------------
-- The untyped λ-calculus, using de Bruijn indices.

_/₀_ : ∀ {n} → Lam (suc n) → Lam n → Lam n
e /₀ a = e /ₛ (subₛ a)

-- etc.

open 𝕊.VarSubst renaming (_↑ to _↑ᵣ; wk to wkᵣ; id to idᵣ) hiding (sub; _⊙_)
open 𝕊.TermSubst tmSubst renaming (_/Var_ to _/ᵣ_)

_/₀-lemma_/ₛ_ : ∀ {m n} e a (σ : Sub Lam m n) →
            (e /₀ a) /ₛ σ ≡ (e /ₛ (σ ↑ₛ)) /₀ (a /ₛ σ)
e /₀-lemma a /ₛ σ = begin
  (e /₀ a) /ₛ σ
   ≡⟨ sub-commutes e ⟩
  (e /ₛ (σ ↑ₛ)) /₀ (a /ₛ σ)
   ∎
      
_/₀-lemma_/ᵣ_ : ∀ {m n} e a (ρ : Sub Fin m n) →
            (e /₀ a) /ᵣ ρ ≡ (e /ᵣ (ρ ↑ᵣ)) /₀ (a /ᵣ ρ)
e /₀-lemma a /ᵣ ρ = begin
  (e /₀ a) /ᵣ ρ
   ≡⟨ sub-commutes-with-renaming e ⟩
  (e /ᵣ (ρ ↑ᵣ)) /₀ (a /ᵣ ρ)
   ∎

--------------------------------------------------------------------
-- top-level beta reduction rule

data _→β_ {n} : Rel (Lam n) ℓ₀ where

  _β_ : ∀ e a → ((ƛ e) • a) →β (e /₀ a)
  
{- admissible rules: closure under renaming, substitution -}

_β/ᵣ_ : ∀ {m n} {e₁ e₂} (r : e₁ →β e₂) (ρ : Sub Fin m n) →
        (e₁ /ᵣ ρ) →β (e₂ /ᵣ ρ)
(e β a) β/ᵣ ρ rewrite e /₀-lemma a /ᵣ ρ = _ β _


_β/_ : ∀ {m n} {e₁ e₂} (r : e₁ →β e₂) (σ : Sub Lam m n) →
       (e₁ /ₛ σ) →β (e₂ /ₛ σ)
(e β a) β/ σ rewrite e /₀-lemma a /ₛ σ = _ β _


---------------------------------------------------------------------------
-- full reduction as the least pre-congruence containing beta

data _↠β_ {n} : Rel (Lam n) ℓ₀ where

  [_]ᵥ : ∀ j →
         ---------------
         ---------------
         [ j ]ᵥ ↠β [ j ]ᵥ

  ƛ    : ∀ {e₁ e₂}       (e : e₁ ↠β e₂) →
         -------------------------------
         (ƛ e₁) ↠β (ƛ e₂)

  _•_  : ∀ {e₁ e₂ a₁ a₂} (e : e₁ ↠β e₂) (r : a₁ ↠β a₂) →
         -----------------------------------------------
         (e₁ • a₁) ↠β (e₂ • a₂)

  _∗_  : ∀ {e₁ e e₂}     (l : e₁ ↠β e) (r : e ↠β e₂) → 
         ---------------------------------------------
         e₁ ↠β e₂

  [_]β : ∀ {e₁ e₂} (r : e₁ →β e₂) → e₁ ↠β e₂
  

-- reflexivity

↠β-refl : ∀ {n} (e : Lam n) → e ↠β e
↠β-refl [ j ]ᵥ  = [ j ]ᵥ
↠β-refl (ƛ e)   = ƛ (↠β-refl e)
↠β-refl (e • a) = ↠β-refl e • ↠β-refl a

_[↠β]_ : ∀ {m n} (σ σ' : Sub Lam m n) → Set
_[↠β]_ = Pointwise _↠β_

[↠β]ₛ : ∀ {m n} {σ : Sub Lam m n} → σ [↠β] σ
[↠β]ₛ = [ ↠β-refl _ ]

-- one step

β∗ : ∀ {n} {e} {a} → ((ƛ {n} e) • a) ↠β (e /₀ a)
β∗ = [ _ β _ ]β


---------------------------------------------------------------------------
-- {- weak-head reduction -}

{- worker/wrapper definition: distinguish zero- from many-step instances -}

data _↠w_ {n} : Rel (Lam n) ℓ₀ 
data _↠⁺w_ {n} : Rel (Lam n) ℓ₀ 

data _↠w_ where

  [] : ∀ {e} → e ↠w e

  _⁻ : ∀ {e₁ e₂} (r : e₁ ↠⁺w e₂) → e₁ ↠w e₂

data _↠⁺w_ where

  _•_ : ∀ {e₁ e₂} (w : e₁ ↠⁺w e₂) a → (e₁ • a) ↠⁺w (e₂ • a)

  _∗_ : ∀ {e₁ e e₂} (l : e₁ ↠⁺w e) (r : e ↠⁺w e₂) → e₁ ↠⁺w e₂

  [_]β : ∀ {e₁ e₂} (r : e₁ →β e₂) → e₁ ↠⁺w e₂
  
{- smart constructors: one-step injection; β instance -}

β[_]ʷ : ∀ {n} {e₁ e₂ : Lam n} → e₁ →β e₂ → e₁ ↠w e₂
β[ r ]ʷ = [ r ]β ⁻

βʷ : ∀ {n} {e} {a} → ((ƛ {n} e) • a) ↠w (e /₀ a)
βʷ = β[ _ β _ ]ʷ

{- admissible rules: transitivities; congruences -}

_w⁺∗w_ : ∀ {n} {e₁ e e₂ : Lam n} → e₁ ↠⁺w e → e ↠w e₂ → e₁ ↠⁺w e₂
l w⁺∗w [] = l
l w⁺∗w (r ⁻) = l ∗ r

_w∗w⁺_ : ∀ {n} {e₁ e e₂ : Lam n} → e₁ ↠w e → e ↠⁺w e₂ → e₁ ↠⁺w e₂
[] w∗w⁺ r = r
(l ⁻) w∗w⁺ r = l ∗ r

_∗w_ : ∀ {n} {e₁ e e₂ : Lam n} → e₁ ↠w e → e ↠w e₂ → e₁ ↠w e₂
l ∗w [] = l
l ∗w (r ⁻) = (l w∗w⁺ r)⁻

_•w_ : ∀ {n} {e₁ e₂ : Lam n} (h : e₁ ↠w e₂) a → (e₁ • a) ↠w (e₂ • a)
[] •w a = []
(h ⁻) •w a = (h • a)⁻

{- admissible rules: closure under renaming, substitution -}

_w⁺/ᵣ_ : ∀ {m n} {e₁ e₂} (w : e₁ ↠⁺w e₂) (ρ : Sub Fin m n) →
        (e₁ /ᵣ ρ) ↠⁺w (e₂ /ᵣ ρ)
_w/ᵣ_  : ∀ {m n} {e₁ e₂} (w : e₁ ↠w e₂) (ρ : Sub Fin m n) →
         (e₁ /ᵣ ρ) ↠w (e₂ /ᵣ ρ)

(w • a)   w⁺/ᵣ ρ = (w w⁺/ᵣ ρ) • (a /ᵣ ρ)
(wl ∗ wr) w⁺/ᵣ ρ = (wl w⁺/ᵣ ρ) ∗ (wr w⁺/ᵣ ρ)
[ r ]β    w⁺/ᵣ ρ = [ r β/ᵣ ρ ]β

[] w/ᵣ ρ   = []
(w ⁻) w/ᵣ ρ = (w w⁺/ᵣ ρ) ⁻

_w⁺/_ : ∀ {m n} {e₁ e₂} (w : e₁ ↠⁺w e₂) (σ : Sub Lam m n) →
        (e₁ /ₛ σ) ↠⁺w (e₂ /ₛ σ)
_w/_  : ∀ {m n} {e₁ e₂} (w : e₁ ↠w e₂) (σ : Sub Lam m n) →
        (e₁ /ₛ σ) ↠w (e₂ /ₛ σ)

(w • a)   w⁺/ σ = (w w⁺/ σ) • (a /ₛ σ)
(wl ∗ wr) w⁺/ σ = (wl w⁺/ σ) ∗ (wr w⁺/ σ)
[ r ]β    w⁺/ σ = [ r β/ σ ]β

[]    w/ σ = []
(w ⁻) w/ σ = (w w⁺/ σ)⁻

{- standard reduction, mutual with internal reduction -}

data _↠i_ {n} : Rel (Lam n) ℓ₀ 
data _↠s_ {n} : Rel (Lam n) ℓ₀

data _↠i_ where

  [_]ᵥ : ∀ j → [ j ]ᵥ ↠i [ j ]ᵥ

  ƛ    : ∀ {e₁ e₂} (i : e₁ ↠s e₂) → ƛ e₁ ↠i ƛ e₂

  _•_ : ∀ {e₁ e₂ a₁ a₂} (l : e₁ ↠i e₂) (r : a₁ ↠s a₂) →
        (e₁ • a₁) ↠i (e₂ • a₂)

data _↠s_ where

  _∗_ : ∀ {e₁ e e₂} (h : e₁ ↠w e) (i : e ↠i e₂) → e₁ ↠s e₂

{- reflexivity, and smart constructors: injections from w/i to s -}

ⁱ[_]ˢ : ∀ {n} {e₁ e₂ : Lam n} → e₁ ↠i e₂ → e₁ ↠s e₂
ⁱ[ i ]ˢ = [] ∗ i

[]ⁱ : ∀ {n} {e : Lam n} → e ↠i e
[]ˢ : ∀ {n} {e : Lam n} → e ↠s e

[]ⁱ {e = [ _ ]ᵥ }  = [ _ ]ᵥ
[]ⁱ {e = ƛ _ }     = ƛ []ˢ
[]ⁱ {e = _ • _}    = []ⁱ • []ˢ

[]ˢ = ⁱ[ []ⁱ ]ˢ

ʷ[_]ˢ : ∀ {n} {e₁ e₂ : Lam n} → e₁ ↠w e₂ → e₁ ↠s e₂
ʷ[ w ]ˢ = w ∗ []ⁱ 

{- smart constructors: one-step, ↠w-prefixing, congruences -}

β[_]ˢ : ∀ {n} {e₁ e₂ : Lam n} (r : e₁ →β e₂) → e₁ ↠s e₂
β[ r ]ˢ = ʷ[ β[ r ]ʷ ]ˢ

βˢ : ∀ {n} {e} {a} → ((ƛ {n} e) • a) ↠s (e /₀ a)
βˢ = β[ _ β _ ]ˢ

_w∗s_ : ∀ {n} {e₁ e e₂ : Lam n} (w : e₁ ↠w e) (s : e ↠s e₂) → e₁ ↠s e₂
wl w∗s (wr ∗ i) = (wl ∗w wr) ∗ i

[_]ˢ : ∀ {n} (j : Fin n) → [ j ]ᵥ ↠s [ j ]ᵥ
[ j ]ˢ = ⁱ[ [ j ]ᵥ ]ˢ

_•s_ : ∀ {n} {e₁ e₂ a₁ a₂ : Lam n} (l : e₁ ↠s e₂) (r : a₁ ↠s a₂) →
       (e₁ • a₁) ↠s (e₂ • a₂)
(w ∗ i) •s s = (w •w _) ∗ (i • s)

ƛs : ∀ {n} {e₁ e₂ : Lam (suc n)} (s : e₁ ↠s e₂) → (ƛ e₁) ↠s (ƛ e₂)
ƛs s = ⁱ[ ƛ s ]ˢ

{- smart constructors: closure under renaming, substitution -}

_[↠s]_ : ∀ {m n} (ρ σ : Sub Lam m n) → Set
_[↠s]_ = Pointwise _↠s_

[↠s]ₛ : ∀ {m n} {σ : Sub Lam m n} → σ [↠s] σ
[↠s]ₛ = [ []ˢ ]

[sub_↠s] : ∀ {n} {e₁ e₂ : Lam n} (e : e₁ ↠s e₂) → (sub e₁) [↠s] (sub e₂)
[sub e ↠s] = e ∷ [↠s]ₛ

_i/ᵣ_ : ∀ {m n} {e₁ e₂} (i : e₁ ↠i e₂) (ρ : Sub Fin m n) →
       (e₁ /ᵣ ρ) ↠i (e₂ /ᵣ ρ)
_s/ᵣ_  : ∀ {m n} {e₁ e₂} (s : e₁ ↠s e₂) (ρ : Sub Fin m n) →
        (e₁ /ᵣ ρ) ↠s (e₂ /ᵣ ρ)

[ j ]ᵥ  i/ᵣ ρ = [ ρ ⟦ j ⟧ ]ᵥ
ƛ s     i/ᵣ ρ = ƛ (s s/ᵣ (ρ ↑ᵣ))
(i • s) i/ᵣ ρ = (i i/ᵣ ρ) • (s s/ᵣ ρ)

(w ∗ i) s/ᵣ ρ = (w w/ᵣ ρ) ∗ (i i/ᵣ ρ)

_[↠s]↑ :  ∀ {m n} {σ₁ σ₂ : Sub Lam m n} (σ : σ₁ [↠s] σ₂) →
           (σ₁ ↑) [↠s] (σ₂ ↑)
σ [↠s]↑ = []ˢ ∷ map⁺ (_s/ᵣ wkᵣ) σ

_i/_ : ∀ {m n} {e₁ e₂} (i : e₁ ↠i e₂)
       {σ₁ σ₂ : Sub Lam m n} (σ : σ₁ [↠s] σ₂) →
       (e₁ /ₛ σ₁) ↠s (e₂ /ₛ σ₂)
_s/_ : ∀ {m n} {e₁ e₂} (s : e₁ ↠s e₂)
       {σ₁ σ₂ : Sub Lam m n} (σ : σ₁ [↠s] σ₂) →
       (e₁ /ₛ σ₁) ↠s (e₂ /ₛ σ₂)

[ j ]ᵥ  i/ σ = σ at j
ƛ s     i/ σ = ƛs (s s/ (σ [↠s]↑))
(i • s) i/ σ = (i i/ σ) •s (s s/ σ)

(w ∗ i) s/ σ = (w w/ _) w∗s (i i/ σ)

-- finally, closure under single parallel substitution

_s/₀_ : ∀ {n} {e₁ e₂} {a₁ a₂ : Lam n} (e : e₁ ↠s e₂) (a : a₁ ↠s a₂) →
       (e₁ /₀ a₁) ↠s (e₂ /₀ a₂)

e s/₀ a = e s/ [sub a ↠s]

{- admissible rules: semi-standardisation -}

{- one-step -}
_i∗β_ : ∀ {n} {e₁ e e₂ : Lam n} (i : e₁ ↠i e) (r : e →β e₂) → e₁ ↠s e₂
(ƛ r • s) i∗β (_ β _) = βʷ w∗s (r s/₀ s)

{- many-step: worker/wrapper -}

_i∗w_ : ∀ {n} {e₁ e e₂ : Lam n} (i : e₁ ↠i e) (w : e ↠w e₂) → e₁ ↠s e₂
_i∗w⁺_ : ∀ {n} {e₁ e e₂ : Lam n} (i : e₁ ↠i e) (w : e ↠⁺w e₂) → e₁ ↠s e₂

{- wrapper -}
i i∗w [] = ⁱ[ i ]ˢ
i i∗w (w ⁻) = i i∗w⁺ w

{- worker -}
i i∗w⁺ (r ∗ s) with w ∗ i ← i i∗w⁺ r = w w∗s (i i∗w⁺ s)

(i • r) i∗w⁺ (w • _) = (i i∗w⁺ w) •s r

i i∗w⁺ [ r ]β        = i i∗β r

{- admissible rules: transitivities -}

_∗i_  : ∀ {n} {e₁ e e₂ : Lam n} (l : e₁ ↠i e) (r : e ↠i e₂) → e₁ ↠i e₂
_∗s_  : ∀ {n} {e₁ e e₂ : Lam n} (l : e₁ ↠s e) (r : e ↠s e₂) → e₁ ↠s e₂

[ j ]ᵥ ∗i i        = i
ƛ l ∗i ƛ r         = ƛ (l ∗s r)
(m • p) ∗i (n • q) = (m ∗i n) • (p ∗s q)

(wl ∗ il) ∗s (wr ∗ ir) with w ∗ i ← il i∗w wr = (wl ∗w w) ∗ (i ∗i ir)

{- additional transitivities -}

_s∗i_ : ∀ {n} {e₁ e e₂ : Lam n} (s : e₁ ↠s e) (i : e ↠i e₂) → e₁ ↠s e₂
_i∗s_ : ∀ {n} {e₁ e e₂ : Lam n} (i : e₁ ↠i e) (s : e ↠s e₂) → e₁ ↠s e₂

(w ∗ il) s∗i ir = w ∗ (il ∗i ir)

il i∗s (w ∗ ir) = (il i∗w w) s∗i ir

{- the Standardisation Theorem -}

weak-head-standardisation :
  ∀ {n} {e₁ e₂ : Lam n} (r : e₁ ↠β e₂) → e₁ ↠s e₂

weak-head-standardisation [ j ]ᵥ  = [ j ]ˢ
weak-head-standardisation (ƛ r)   = ƛs (weak-head-standardisation r)
weak-head-standardisation (r • s) =
  (weak-head-standardisation r) •s (weak-head-standardisation s)
weak-head-standardisation (r ∗ s) =
  (weak-head-standardisation r) ∗s (weak-head-standardisation s)
weak-head-standardisation [ r ]β  = β[ r ]ˢ

---------------------------------------------------------------------------
{- fin -}
