---------------------------------------------------------------------------
-- The Church-Rosser Theorem for untyped λ-calculus

{-# OPTIONS --without-K --safe #-}

module super-Church-Rosser where

open import Data.Bool.Base using (Bool; true; false; not; T)
open import Data.Nat.Base  using (zero; suc)
open import Data.Fin.Base  using (Fin; zero; suc)
open import Data.Product   using (∃-syntax; _,_; _×_)

open import Data.Vec.Base using (_∷_; map)
  renaming ([] to ε; lookup to _⟦_⟧)
open import Data.Vec.Relation.Binary.Pointwise.Inductive
  renaming ([] to ε∼ε; refl to [_])
  hiding (map)

open import Level renaming (zero to ℓ₀) hiding (suc)

open import Relation.Binary
open import Relation.Binary.PropositionalEquality as ℙ
  using (_≡_; refl; sym; cong; cong₂)
open ℙ.≡-Reasoning

open import Relation.Unary using (Pred)

open import Data.Fin.Substitution as 𝕊
open import Data.Fin.Substitution.Lemmas as 𝕃

---------------------------------------------------------------------------
-- The untyped λ-calculus, using de Bruijn indices.

open import Data.Fin.Substitution.Example as 𝔼
  renaming (Tm to Lam; var to [_]ᵥ; _·_ to _•_;
            _/_ to _/ₛ_; _↑ to _↑ₛ; _↑⋆_ to _↑ₛ⋆_;
            weaken to weakenₛ; wk to wkₛ;
            sub to subₛ; id to idₛ; _⊙_ to _⊙ₛ_)

open 𝕊.VarSubst renaming (_↑ to _↑ᵣ; wk to wkᵣ; id to idᵣ) hiding (sub; _⊙_)
open 𝕊.TermSubst tmSubst renaming (_/Var_ to _/ᵣ_)

-- some auxiliary stuff

_/₀_ : ∀ {n} → Lam (suc n) → Lam n → Lam n
e /₀ a = e /ₛ (subₛ a)

_/₀-lemma_/ₛ_ : ∀ {m n} e a (σ : Sub Lam m n) →
            (e /₀ a) /ₛ σ ≡ (e /ₛ (σ ↑ₛ)) /₀ (a /ₛ σ)
e /₀-lemma a /ₛ σ = begin
  (e /₀ a) /ₛ σ
   ≡⟨ sub-commutes e ⟩
  (e /ₛ (σ ↑ₛ)) /₀ (a /ₛ σ)
   ∎
      
_/₀-lemma_/ᵣ_ : ∀ {m n} e a (ρ : Sub Fin m n) →
            (e /₀ a) /ᵣ ρ ≡ (e /ᵣ (ρ ↑ᵣ)) /₀ (a /ᵣ ρ)
e /₀-lemma a /ᵣ ρ = begin
  (e /₀ a) /ᵣ ρ
   ≡⟨ sub-commutes-with-renaming e ⟩
  (e /ᵣ (ρ ↑ᵣ)) /₀ (a /ᵣ ρ)
   ∎

---------------------------------------------------------------------------
-- super-parallel reduction rules
-- (Welch 1975; Lévy 1977; Aczel 1979; Klop 1980; van Raamsdonk 1993)

data _⇶_ {n} :  Rel (Lam n) ℓ₀ where

  [_]ᵥ : ∀ j →
         ---------------
         ---------------
         [ j ]ᵥ ⇶ [ j ]ᵥ

  ƛ    : ∀ {e₁ e₂}       (e : e₁ ⇶   e₂) →
         -------------------------------
         (ƛ e₁) ⇶ (ƛ e₂)

  _•_  : ∀ {e₁ e₂ a₁ a₂} (e : e₁ ⇶   e₂) (a : a₁ ⇶ a₂) →
         ----------------------------------------------
         (e₁ • a₁) ⇶ (e₂ • a₂)

  _β_  : ∀ {e₁ e₂ a₁ a₂} (e : e₁ ⇶ ƛ e₂) (a : a₁ ⇶ a₂) →
         ----------------------------------------------
         (e₁ • a₁) ⇶ (e₂ /₀ a₂)

-- reflexivity

⇶-refl : ∀ {n} (e : Lam n) → e ⇶ e
⇶-refl [ j ]ᵥ  = [ j ]ᵥ
⇶-refl (ƛ e)   = ƛ (⇶-refl e)
⇶-refl (e • a) = ⇶-refl e • ⇶-refl a

_⇶ₛ_ : ∀ {m n} (σ₁ σ₂ : Sub Lam m n) → Set
_⇶ₛ_ = Pointwise _⇶_

[⇶ₛ] : ∀ {m n} {σ : Sub Lam m n} → σ ⇶ₛ σ
[⇶ₛ] = [ ⇶-refl _ ]

[sub_⇶ₛ] : ∀ {n} {e₁ e₂ : Lam n} (e : e₁ ⇶ e₂) → (sub e₁) ⇶ₛ (sub e₂)
[sub e ⇶ₛ] = e ∷ [⇶ₛ]

-- closure under parallel substitution

_⇶/ᵣ_ : ∀ {m n} {e₁ e₂} (e : e₁ ⇶ e₂) (ρ : Sub Fin m n) →
        (e₁ /ᵣ ρ) ⇶ (e₂ /ᵣ ρ)

[ j ]ᵥ  ⇶/ᵣ ρ = [ ρ ⟦ j ⟧ ]ᵥ
ƛ e     ⇶/ᵣ ρ = ƛ (e ⇶/ᵣ (ρ ↑ᵣ))
(e • a) ⇶/ᵣ ρ = (e ⇶/ᵣ ρ) • (a ⇶/ᵣ ρ)
_β_ {e₁ = e₁} {e₂ = e₂} {a₁ = a₁} {a₂ = a₂} e a ⇶/ᵣ ρ = beta
  where
    beta : ((e₁ • a₁) /ᵣ ρ) ⇶ ((e₂ /₀ a₂) /ᵣ ρ) 
    beta rewrite e₂ /₀-lemma a₂ /ᵣ ρ = (e ⇶/ᵣ ρ) β (a ⇶/ᵣ ρ)

_⇶ₛ↑ : ∀ {m n} {σ₁ σ₂ : Sub Lam m n} (σ : σ₁ ⇶ₛ σ₂) →
        (σ₁ ↑ₛ) ⇶ₛ (σ₂ ↑ₛ)
σ ⇶ₛ↑ = (⇶-refl _) ∷ map⁺ (_⇶/ᵣ wkᵣ) σ

_⇶/ₛ_ : ∀ {m n} {e₁ e₂} (e : e₁ ⇶ e₂)
        {σ₁ σ₂ : Sub Lam m n} (σ : σ₁ ⇶ₛ σ₂) →
        (e₁ /ₛ σ₁) ⇶ (e₂ /ₛ σ₂)

[ j ]ᵥ ⇶/ₛ σ  = lookup σ j
ƛ e    ⇶/ₛ σ  = ƛ (e ⇶/ₛ (σ ⇶ₛ↑))
(e • a) ⇶/ₛ σ = (e ⇶/ₛ σ) • (a ⇶/ₛ σ)
_⇶/ₛ_ (_β_ {e₁ = e₁} {e₂ = e₂} {a₁ = a₁} {a₂ = a₂} e a)
      {σ₁ = σ₁} {σ₂ = σ₂} σ = beta
  where
    beta : ((e₁ • a₁) /ₛ σ₁) ⇶ ((e₂ /₀ a₂) /ₛ σ₂) 
    beta rewrite e₂ /₀-lemma a₂ /ₛ σ₂ = (e ⇶/ₛ σ) β (a ⇶/ₛ σ)

_⇶/₀⇶_ : ∀ {n} {e₁ e₂ : Lam (suc n)} {a₁ a₂}
          (e : e₁ ⇶ e₂) (a : a₁ ⇶ a₂) →
          (e₁ /₀ a₁) ⇶ (e₂ /₀ a₂)
         
e ⇶/₀⇶ a = e ⇶/ₛ [sub a ⇶ₛ]


---------------------------------------------------------------------------
-- complete super-development rules
-- (Welch 1975; Lévy 1977; Aczel 1979; Klop 1980; van Raamsdonk 1993) 

-- first, an important observer

ƛ? : ∀ {n} (e : Lam n) → Bool
ƛ? [ _ ]ᵥ  = false
ƛ? (ƛ _)   = true
ƛ? (_ • _) = false

-- finally 

data _⇶ᶜ_ {n} :  Rel (Lam n) ℓ₀ where

  [_]ᵥ    : ∀ j → 
            ----------------
            ----------------
            [ j ]ᵥ ⇶ᶜ [ j ]ᵥ

  ƛ       : ∀ {e₁ e₂} (e : e₁ ⇶ᶜ e₂) →
            ------------------------
            (ƛ e₁) ⇶ᶜ (ƛ e₂)

  _•[_]_  : ∀ {e₁ e₂ a₁ a₂}
              (e : e₁ ⇶ᶜ e₂) (notƛ : T (not (ƛ? e₂)))
              (a : a₁ ⇶ᶜ a₂) →
            -----------------------------------------------
            (e₁ • a₁) ⇶ᶜ (e₂ • a₂)

  _β_     : ∀ {e₁ e₂ a₁ a₂} (e : e₁ ⇶ᶜ ƛ e₂) (a : a₁ ⇶ᶜ a₂) →
            -----------------------------------------------
            (e₁ • a₁) ⇶ᶜ (e₂ /₀ a₂)

-- every complete superdevelopment is a superparallel reduction

⇶ᶜ-⇶ : ∀ {n} {e₁ e₂ : Lam n} (d : e₁ ⇶ᶜ e₂) → e₁ ⇶ e₂

⇶ᶜ-⇶ [ j ]ᵥ       = [ j ]ᵥ
⇶ᶜ-⇶ (ƛ l)        = ƛ (⇶ᶜ-⇶ l)
⇶ᶜ-⇶ (l •[ _ ] r) = ⇶ᶜ-⇶ l • ⇶ᶜ-⇶ r
⇶ᶜ-⇶ (l β r)      = ⇶ᶜ-⇶ l β ⇶ᶜ-⇶ r

--------------------------------------------------------------------------
-- triangle lemma (Takahashi 1989/95; adapted to 'super' case)

_⇶-⇶ᶜ-⇶_ : ∀ {n} {e₀ e₁ eᵈ : Lam n}
             (r : e₀ ⇶ e₁) (d : e₀ ⇶ᶜ eᵈ) → e₁ ⇶ eᵈ

[ _ ]ᵥ    ⇶-⇶ᶜ-⇶ [ _ ]ᵥ             = [ _ ]ᵥ

(ƛ r)     ⇶-⇶ᶜ-⇶ (ƛ l)             = ƛ (r ⇶-⇶ᶜ-⇶ l)

(rl • rr) ⇶-⇶ᶜ-⇶ (dl •[ notƛ ] dr) = rl ⇶-⇶ᶜ-⇶ dl • rr ⇶-⇶ᶜ-⇶ dr

(rl • rr) ⇶-⇶ᶜ-⇶ (dl β dr)         = (rl ⇶-⇶ᶜ-⇶ dl) β (rr ⇶-⇶ᶜ-⇶ dr)

(rl β rr) ⇶-⇶ᶜ-⇶ (dl •[ notƛ ] dr) with ƛ r ← rl ⇶-⇶ᶜ-⇶ dl | () ← notƛ

(rl β rr) ⇶-⇶ᶜ-⇶ (dl β dr)         with ƛ r ← rl ⇶-⇶ᶜ-⇶ dl
                                     = r ⇶/₀⇶ (rr ⇶-⇶ᶜ-⇶ dr)

---------------------------------------------------------------------------
-- Church-Rosser (Takahashi 1989/95; adapted)

-- superdevelopments exist

∃⇶ᶜ : ∀ {n} (e : Lam n) → ∃[ eᶜ ] e ⇶ᶜ eᶜ

∃⇶ᶜ [ j ]ᵥ                         = [ j ]ᵥ , [ j ]ᵥ
∃⇶ᶜ (ƛ e)   with (e , eᶜ) ← ∃⇶ᶜ e = ƛ e , ƛ eᶜ
∃⇶ᶜ (e • a) with (r , eᶜ) ← ∃⇶ᶜ e | (s , aᶜ) ← ∃⇶ᶜ a | r
... | ƛ r        = r /₀ s , eᶜ β aᶜ
... | r@([ _ ]ᵥ) =  r • s , eᶜ •[ _ ] aᶜ
... | r@(_ • _)  =  r • s , eᶜ •[ _ ] aᶜ

-- so two appeals to the triangle lemma clinch it

_⇶-⇶-⇶_ : ∀ {n} {e e₁ e₂ : Lam n} (l : e ⇶ e₁) (r : e ⇶ e₂) →
            ∃[ eᶜ ] e₁ ⇶ eᶜ × e₂ ⇶ eᶜ
        
_⇶-⇶-⇶_ {e = e} e₁ e₂ with eᶜ , dᶜ ← ∃⇶ᶜ e =
           eᶜ , e₁ ⇶-⇶ᶜ-⇶ dᶜ , e₂ ⇶-⇶ᶜ-⇶ dᶜ

--------------------------------------------------------------------------
{- fin -}
